#!/usr/bin/env python
#################################################################################################
#            module: pGQL.py
#           authors: Ruben Schilling (ruben.schilling@molgen.mpg.de)
#                    Ivan G. Costa (igcf@cin.ufpe.br)
#                    Alexander Schliep (alexander.schliep@molgen.mpg.de)
#          version:  0.1
#      description:  This executable file enables the user to visually
#                    define linear HMM (query) models on time course data.
#
#     Copyright (C)  2009, 2010 Ruben Schilling, Ivan G. Costa and Alexander Schliep
#
#          Contact:  ruben.schilling@molgen.mpg.de
#
#      Information:  http://ghmm.org/gql
#
#   pGQL is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   pGQL is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pGQL; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#   or see <http://www.gnu.org/licenses/>.
#################################################################################################

from Tkinter import *
import tkFileDialog
import tkMessageBox
import tkFont
import string
from math import *
from GQLUtil import *
from re import *
import sets
import copy
from ghmm import * 
import GQL
import numpy as Numeric
import sys
import os
#------------------------------------------------------------------------------------------------------
#******************************************************************************************************
#------------------------------------------------------------------------------------------------------
class GQLQuery:
    """ A GQLQuery is a container (models) for the query
        formulation and the results.  It keeps tracks of listeners (views) for
        result updates.

        Controllers use the set... functions.

        self.allRanked contains all the sequence ids ranked by decreasing
        likelihood (NOTE: 0.0 likelihood is understood to be -infinity)
        Stupid convention in GHMM

        self.queryResult contains the subset of sequence ids ranked
        above (i.e. with lower ranks) than the cutoff.

    """

    def __init__(self):
        self.queryLength = None

        self.allRanked = None # All the sequences ranked by the HMM
        self.queryResult = sets.ImmutableSet() # The displayed set
        self.rankCutoff = 0 # Larger cutoff number == lower rank
        # Rank from 0 to n-1
        self.maxRank = 0
        self.data = None

        self.listeners = []
        self.shmm = None

        self.cyclic_prob = 0.5

        self.likelihood = None
        self.viterbiPaths = []
        self.queryModified = 0
        self.missingRate = 0.0001 # assume a low probabilitie of missing data as default ... 
        self.data = None
        


    def newQuery(self, queryLength, mean, variance, duration, cyclic,densityType=[]):
        """ queryLength    number of emitting states
            mean           vector of mean values for queryLength states
            variance       vector of variance vales for queryLength states
            cyclic         indicator wheter the whole HMM cycles
        """
        if self.queryLength is not None:
            self.shmm = None
            del(self.mean)
            del(self.variance)
            del(self.duration)
            del(self.densityType)

        print mean, variance, duration, densityType
            
        self.queryLength = queryLength
        self.mean = copy.copy(mean)
        self.variance = copy.copy(variance)
        self.duration = copy.copy(duration)
        self.cyclic = cyclic
        if densityType == []: # normal emission by default           
          self.densityType = [ghmmwrapper.normal]*(queryLength+1)
        else:
          self.densityType = copy.copy(densityType)

        if (self.data != None):
            self.missingRate = self.data.missingRate
        self.newSHMM()
        self.runQuery()


    def openQuery(self, fileName, fileType):        
        if (fileType == 'xml'):
          model = self.readXML(fileName)  
        else:
          model = self.readSHMM(fileName)
        if model != None:
          del self.shmm
          self.shmm = model
          self.loadQuery(model)


    def loadQuery(self,model):
        self.shmm = model
        self.mean = []
        self.variance = []
        self.duration = []
        self.densityType = []

        self.queryLength = self.shmm.N - 1 # Correct for 'end' state
        
        for i in xrange(self.queryLength):
            if isinstance(self.shmm,ContinuousMixtureHMM):
                param = self.shmm.getEmission(i,0)
                if ((param[0] == ghmmwrapper.normal) or
                    (param[0] == ghmmwrapper.uniform)):
                  mu = param[1]
                  sigma = param[2]
                  type = param[0]                  
                else:
                  print 'Emission type not suported'
            elif isinstance(self.shmm,GaussianMixtureHMM):
                (mu, sigma,weight) = self.shmm.getEmission(i,0)
                type = ghmmwrapper.normal

            else:
                (mu, sigma) = self.shmm.getEmission(i)
                type = ghmmwrapper.normal
            self.mean.append(mu)
            self.variance.append(sigma)
            self.densityType.append(type)
            a = self.shmm.getTransition(i,i) # Self-transition
            if( a!=1.0):
              self.duration.append(1.0 / (1.0 - a))
            else:
              self.duration.append(9999.99) # big self duration!!!
              
        # check this code 
        if self.shmm.getTransition(self.queryLength-1,0) > 0.0:
            self.cyclic = 1
        else:
            self.cyclic = 0

        # rebuild the model, now with missing data emission ...
        self.newSHMM()
        self.runQuery()


    def saveQuery(self, fileName):
        self.writeSHMM(fileName)


    def setData(self, data):
        """ We presume everything is shown at this point. That means
            rankCutoff equals maxRank and the queryResult is the whole set
        """
        self.data = data

        # because of the missing data rate, we have to rebuild the models
        self.missingRate = self.data.missingRate
        self.newSHMM()
        
        
        self.maxRank = len(self.data)
        self.rankCutoff = self.maxRank
        self.allRanked  = self.rankByLikelihood()
    
        #show = self.allRanked[0:self.rankCutoff]
        show = self.allRanked
        hide = []
        self.queryResult = sets.ImmutableSet(show)
        #print "setData", self.queryResult, self.allRanked, self.maxRank, self.rankCutoff
        self.runQuery()
        
    
    def setRankCutoff(self,newCutoff): 
        """ Set the rank cutoff to the newCutoff """
        if self.data == None: # Cant have cutoff, if we dont have data
            return
    
        if newCutoff < self.rankCutoff: # Get more specific
            show = []
            hide = self.allRanked[newCutoff:self.rankCutoff]
            self.queryResult = self.queryResult.difference(sets.ImmutableSet(hide))
    
        elif newCutoff > self.rankCutoff: # Get less specific
            show = self.allRanked[self.rankCutoff:newCutoff]
            hide = []
            self.queryResult = self.queryResult.union(sets.ImmutableSet(show))
    
        else: # Nothing changed, nothing to do
            return
    
        self.updateResult(show,hide)
        self.rankCutoff = newCutoff
    
    
    def runQuery(self):
        self.allRanked = self.rankByLikelihood()
        newQueryResult = sets.ImmutableSet(self.allRanked[0:self.rankCutoff])
        hide = self.queryResult.difference(newQueryResult)
        show = newQueryResult.difference(self.queryResult)
        self.queryResult = newQueryResult
        self.updateResult(show, hide)
        self.queryModified = 1  

    def updateResult(self, showProfiles, hideProfiles):
        for l in self.listeners:
            l.update(showProfiles, hideProfiles)

    def addListener(self, listener):
        self.listeners.append(listener)

    def removeListener(self, listener):
        self.listeners.remove(listener)

    def Result(self):
        return self.queryResult

    def LowestRankedLikelihood(self):
        if self.likelihood != None and self.allRanked != None:
            p = self.allRanked[self.rankCutoff-1]
            return self.likelihood[p]
        else:
            return 0.0 # Should not get here


    def Likelihood(self, profileNr):
        if self.likelihood != None:
            return self.likelihood[profileNr]
        else:
            return 0.0 # Should not get here


    def newSHMM(self):
        del self.shmm
        N = self.queryLength + 1 # Correct for 'end' state

        # Build the transition matrix
        A = Numeric.zeros((N,N), Numeric.float)

        for i in range(N - 1):
            p = 1.0 - 1.0/self.duration[i]
            A[i,i] = p
            A[i,i+1] = 1.0 - p
        A[N-1,N-1] = 1.0

        if self.cyclic:
            A[N-2,0] = A[N-2,N-1]/2
            A[N-2,N-1] =  A[N-2,N-1]/2         
            
        B = Numeric.zeros((N,4,2), Numeric.float)
        for i in range(N - 1):
            B[i,0,0] = self.mean[i]
            B[i,0,1] = -9999.99 # missing data
            B[i,1,0] = self.variance[i]
            B[i,1,1] = 0.0001
            B[i,2,0] = 0.0
            B[i,2,1] = 0.0               
            B[i,3,0] = 1-self.missingRate
            B[i,3,1] = self.missingRate

        B[N-1,0,0] = 9999.99 # END symbol, only produced in last state
        B[N-1,0,1] = -9999.99 # missing data
        B[N-1,1,0] = 0.01
        B[N-1,1,1] = 0.01
        B[N-1,2,0] = 0.0
        B[N-1,2,1] = 0.0
        B[N-1,3,0] = 1
        B[N-1,3,1] = 0        

        
        pi = Numeric.zeros(N, Numeric.float)
        pi[0] = 1.0

        densities = []
        for i in self.densityType:
            densities.append([i,ghmmwrapper.normal])
        densities.append([ghmmwrapper.normal,ghmmwrapper.normal])

        #print A, B, pi, densities
        #del self.shmm
        self.shmm = HMMFromMatrices(Float(),ContinuousMixtureDistribution(Float()),
                                    A, B, pi,densities=densities)
        for i in range(N):
           self.shmm.setMixtureFix(i,[0,1])
        # the last states is also fixed   
        self.shmm.setStateFix(N-1,1)


    def deleteSHMM(self):
        self.freeSHMM()

    def readSHMM(self, filename):
        """ Read a model from a file using text format (not XML).
            Return an instance of class SHMM """
        model = HMMOpen(filename,0)
        return model

    def readXML(self, filename):
        """ Read a model from a file using text format (not XML).
            Return an instance of class SHMM """
        #XXX - latter on add suport for choosing the model
        model = HMMOpenXML(filename,0)
        if( model == None):
          return None
        #model = all[0] # currently only opens first model
        return model

    def writeSHMM(self, filename):
        """ Write the model to a file in text format """
        self.shmm.write(filename)

    def rankByLikelihood(self):
        if self.data is None:
            return []
        seqs = self.data.GHMMProfileSet() # Returns a ghmm SequenceSet object
        self.likelihood = self.shmm.loglikelihoods(seqs)
        indices = range(len(self.likelihood))
        return SortOnItem(indices, self.likelihood)

    def setDuration(self,k,newDuration):
        self.duration[k] = newDuration
        p = 1.0 - 1.0/self.duration[k]

        if k < self.shmm.N - 2:
            self.shmm.setTransition(k,k  ,p)     # self-transition
            self.shmm.setTransition(k,k+1,1-p)   # next state
            #print p
        elif k == self.shmm.N - 2:
            q = 1.0 - p # transition prob to end without cycling
            if self.cyclic == 0:
                self.shmm.setTransition(k, 0, 0.0) # to first query state
                self.shmm.setTransition(k, k, p) # self-transition
                self.shmm.setTransition(k,k+1,q) # to end state
                #print p
            else:
                # 95 % go to first, remainder to end
                self.shmm.setTransition(k, 0, self.cyclic_prob * q) # to first query state
                self.shmm.setTransition(k, k, p) # self-transition
                self.shmm.setTransition(k,k+1,(1.0-self.cyclic_prob) * q) # to end state
        else:
            self.shmm.setTransition(k,k,1.0) # Should we ever get here?
        self.runQuery()

    def setMean(self,k,newMean):
        if ( k < self.shmm.N - 1):
            self.mean[k] = float(newMean)
            if (self.densityType[k] != ghmmwrapper.uniform):
              self.shmm.setEmission(k,0, self.densityType[k], (self.mean[k],self.variance[k],1-self.missingRate,0))
              self.runQuery()

    def setVariance(self,k,newVariance):
        if ( k < self.shmm.N - 1 ):
            self.variance[k] = float(newVariance)
            if (self.densityType[k] != ghmmwrapper.uniform):
              self.shmm.setEmission(k,0, self.densityType[k], (self.mean[k],self.variance[k],1-self.missingRate,0))
              self.runQuery()

    def setDensity(self,k,newDensity):
        if ( k < self.shmm.N - 1 ):
            self.densityType[k] = newDensity
            if (newDensity == ghmmwrapper.uniform):
              self.shmm.setEmission(k,0, self.densityType[k], (50.0,-50.0,1-self.missingRate,0))  
            else:
              self.shmm.setEmission(k,0, self.densityType[k], (self.mean[k],self.variance[k],1-self.missingRate,0))
            self.runQuery()
            
    def setCyclic(self,cyclic):
        self.cyclic = cyclic
        # Update SHMM parameters: last-to-first, last-last, last-end transition

        k = self.queryLength - 1 # indexing from 0
        p = 1.0 - 1.0/self.duration[k]
        if cyclic == 0:
            self.shmm.setTransition(k, 0, 0.0) # to first query state
            self.shmm.setTransition(k,k,p) # self-transition
            self.shmm.setTransition(k, k+1, 1.0 - p) # to end state
        else:
            q = 1.0 - p # transition prob to end without cycling
            # 95 % go to first, remainder to end
            #print "antes", self.shmm.getTransition(k, 0)
            self.shmm.setTransition(k, 0, self.cyclic_prob * q) # to first query state
            #print "depois", self.shmm.getTransition(k, 0)
            self.shmm.setTransition(k, k, p) # self-transition
            self.shmm.setTransition(k,k+1,(1.0-self.cyclic_prob)* q) # to end state
            #print self.shmm.toMatrices()
            #print self.cyclic_prob
        self.runQuery()

    def isModified(self):
            if self.queryModified:
                if self.data != None:
                    self.viterbiPaths = self.shmm.viterbi(self.data.GHMMProfileSet())
                self.queryModified = 0

def SortOnItem(list, sortKey):
    templist = [ (sortKey[item], item) for item in list ]
    templist.sort()
    templist.reverse()
    return [ item[1] for item in templist ]

#------------------------------------------------------------------------------------------------------
#******************************************************************************************************
#------------------------------------------------------------------------------------------------------
  

#------------------------------------------------------------------------------------------------------
#******************************************************************************************************
#------------------------------------------------------------------------------------------------------     
class ToolTip (object):
    """
        This class was posted by a user called 'Fuzzyman' on 2006-07-06 21:29:24 on the website
        http://www.voidspace.org.uk/python/weblog/arch_d7_2006_07_01.shtml
        No copyright was claimed there and neither is here.
    """

    def __init__ (self, widget):
        self.widget = widget
        self.tipwindow = None
        self.id = None
        self.x = self.y = 0

    def showtip (self, text):
        """
            Display text in tooltip window
        """
        
        self.text = text
        if self.tipwindow or not self.text:
            return
        x, y, cx, cy = self.widget.bbox ("insert")
        x = x + self.widget.winfo_rootx () + 25
        y = y + cy + self.widget.winfo_rooty () +25
        self.tipwindow = tw = Toplevel (self.widget)
        tw.wm_overrideredirect (1)
        tw.wm_geometry ("+%d+%d" % (x, y))
        try:
            # For Mac OS
            tw.tk.call ("::tk::unsupported::MacWindowStyle",
                       "style", tw._w,
                       "help", "noActivates")
        except TclError:
            pass
        label = Label (tw, text=self.text, justify=LEFT,
                      background="#ffffe0", relief=SOLID, borderwidth=1,
                      font=("tahoma", "12", "normal"))
        label.pack (ipadx=1)

    def hidetip (self):
        tw = self.tipwindow
        self.tipwindow = None
        if tw:
            tw.destroy ()

    def createToolTip (self, widget, text):
        toolTip = ToolTip (widget)
        def enter (event):
            toolTip.showtip (text)
        def leave (event):
            toolTip.hidetip ()
        widget.bind ('<Enter>', enter)
        widget.bind ('<Leave>', leave)
#------------------------------------------------------------------------------------------------------
#******************************************************************************************************
#------------------------------------------------------------------------------------------------------



#------------------------------------------------------------------------------------------------------
#******************************************************************************************************
#------------------------------------------------------------------------------------------------------
class TimeBox :
    """
        This class contains all necessary information of time box objects in pGQL.
        The class attributes are set via the constructor only.
        The attributes should be retrieved via the classe's getters.
    """
    
    def __init__( self, hidden, canvasID, mean, variance, duration, begin ) :
        self.hidden = hidden        # True iff this box is a spacer, that represents a hidden state in the HMM
        self.canvasID = canvasID    # The GUI's canvas' id of the rectangle representation of this box
        self.mean = mean            # HMM parameter
        self.variance = variance    # HMM parameter
        self.duration = duration    # HMM parameter
        self.begin = begin          # the lefter most time axis coordinate of this box
        self.handles = []           # list of resize handles belonging to this box

    def isHidden( self ) :
        return self.hidden
    #getters
    def getMean( self ):
        return self.mean
    def getVariance( self ) :
        return self.variance
    def getDuration( self ) :
        return self.duration
    def getBegin( self ) :
        return self.begin
    def getCanvasID( self ) :
        # the canvas ID is a unique identifier for a graphical item on the canvas
        return self.canvasID
    #setters
    def setCanvasID( self, canvasID ) :
        # the canvas ID is a unique identifier for a graphical item on the canvas
        self.canvasID=canvasID
#------------------------------------------------------------------------------------------------------
#******************************************************************************************************
#------------------------------------------------------------------------------------------------------
class ResizeHandle :
    """
        Class to store information of the resize handles. The coordinates stored are based on the 
        time/value coordinate system of the time course data. Thus it is not screen positions, that
        are saved here. The coordinates stored represent the center of a resize handle. The display
        function will draw an oval around that center.
        Position can be northEast or any likewise (string) value.
    """
    def __init__( self, x, y, canvasID, associatedTimeBoxID, position ) :
        self.x=x
        self.y=y
        self.canvasID=canvasID # the canvas ID is a unique identifier for a graphical item on the canvas
        self.associatedTimeBoxID=associatedTimeBoxID
        self.position=position

#------------------------------------------------------------------------------------------------------
#******************************************************************************************************
#------------------------------------------------------------------------------------------------------
class Data :
    """
        This class handles the storage of and access to time boxes and -courses as well as query results
    """
    def __init__( self, GUIMaster ) :
        self.timeCourses = []
        self.timeBoxes = []
        self.queryResult = []
        self.GUIMaster = GUIMaster


    
    def readFile( self, filename ) :
        try :
            f = open( filename, 'r' )
            self.timeCourses = [] # old data is replaced
            self.GUIMaster.queryLabel.configure( text="Query result from file: " + filename ) # show the loaded file path
            for line in f :
                self.timeCourses.append( line )
            f.close()
            self.GUIMaster.rankSlider.configure( to = len( self.timeCourses ) ) 
            self.GUIMaster.rankSlider.set( self.GUIMaster.rankSlider.cget( 'to' ) )
            self.queryResult = copy.copy( self.timeCourses ) # when opening a new file the old query becomes obsolete
        except IOError :
            print "Sorry, there was an error reading the file:", filename
    
    
    
    def writeFile( self, fileName,query ) :
        try :
            f = open( fileName, 'w' )
            for i in query.queryResult :
                #cells =  self.timeCourses[i].split('\t')
                #f.writelines( cells[0]+'\n')
                f.writelines( self.timeCourses[i] )
            f.close()
            #print "wrote file ", fileName
        except IOError :
            print "Sorry, there was an error writing the file:", fileName



    def functionalAnalysisQuery( self, query ) :
        """
            Submit the current query result to gProfiler
        """
        import webbrowser
        URL_GOST="http://biit.cs.ut.ee/gprofiler/index.cgi?organism=scerevisiae&r_chr=X&r_start=start&r_end=end&fdr=1&domain_size_type=annotated&term=&significant=1&user_thr=0.05&output=png&prefix=ENTREZGENE_ACC&query="         
        genes = []
        for i in query.queryResult :
            cells =  self.timeCourses[i].split('\t')
            genes.append(cells[0])
        query=URL_GOST+'+'.join(genes)
        webbrowser.open(query)


    
    def addBox( self, timeBoxInstance ) :
        """
            This method should be used to insert a new time box in the ordered list of boxes
            Invariant: An ordered list of time boxes (by their leftermost 'x' screen coordinate)
        """
        self.timeBoxes.append( timeBoxInstance )
        self.timeBoxes.sort( key=TimeBox.getBegin )
        self.recomputeHiddenBoxes()
        
        
    
    def recomputeHiddenBoxes( self ) :
        """
            After drawing, moving, deleting or resizing a timebox this method is invoked.
            It deletes all existing hidden boxes and inserts the new correct hidden boxes.
        """
        # Remove all hidden boxes
        removables = []
        for box in self.timeBoxes :
            if box.hidden :
                removables.append( box ) # collect boxes for removal
        for box in removables :
            self.timeBoxes.remove( box )
        del removables
        # Recompute all hidden boxes
        # Initially the hidden boxes list contains one hidden box spanning the entire time axis
        hiddenBoxIntervals = [ [ 0.5, self.GUIMaster.renderer.timepoints-1 ] ]
        # Split the hidden box and/or remove periods for which boxes exist
        for box in self.timeBoxes :
            for interval in hiddenBoxIntervals :
                if box.begin == interval[ 0 ] and box.begin+box.duration == interval[ 1 ] :
                    hiddenBoxIntervals.remove( interval )
                    break
                if box.begin > interval[ 0 ] and box.begin+box.duration < interval[ 1 ] :
                    split1 = [ interval[ 0 ], box.begin ]
                    split2 = [ box.begin+box.duration, interval[ 1 ] ]
                    hiddenBoxIntervals.remove( interval )
                    hiddenBoxIntervals.append (split1 )
                    hiddenBoxIntervals.append (split2 )
                    break
                if box.begin == interval[ 0 ] and box.begin+box.duration < interval[ 1 ] :
                    split = [ box.begin+box.duration, interval[ 1 ] ]
                    hiddenBoxIntervals.remove( interval )
                    hiddenBoxIntervals.append( split )
                    break
                if box.begin > interval[ 0 ] and box.begin+box.duration == interval[ 1 ] :
                    split = [ interval[ 0 ], box.begin ]
                    hiddenBoxIntervals.remove( interval )
                    hiddenBoxIntervals.append( split )
                    break
        # construct all hidden Time boxes
        for i in hiddenBoxIntervals :
            box = TimeBox(   True, \
                                None, \
                                self.GUIMaster.renderer.highestValue, \
                                self.GUIMaster.renderer.lowestValue, \
                                i[ 1 ]-i[ 0 ], \
                                i[ 0 ] )
            self.timeBoxes.append( box )
        self.timeBoxes.sort( key=TimeBox.getBegin )
    
            
    
    # these methods return parameters for HMM creation
    def getMeanVector( self ) :
        means = []
        for box in self.timeBoxes :
            means.append( box.getMean() )
        return means
    
    
    
    def getVarianceVector( self ) :
        variances = []
        for box in self.timeBoxes :
            variances.append( box.getVariance() )
        return variances
    
    
    
    def getDurationsVector( self ) :
        durations = []
        for box in self.timeBoxes :
            durations.append( box.getDuration() )
        return durations
    
    
    
    def getQueryLength( self ) :
        return len( self.timeBoxes )
#------------------------------------------------------------------------------------------------------
#******************************************************************************************************
#------------------------------------------------------------------------------------------------------

        

#------------------------------------------------------------------------------------------------------
#******************************************************************************************************
#------------------------------------------------------------------------------------------------------
class Interaction :

    def __init__( self, gui ) :
        self.gui = gui
        # We draw the mouse from one position ( x1,y1 ) to another position ( x2,y2 )
        self.tempX1 = self.tempY1 = self.tempX2 = self.tempY2 = 0
        self.allowDraw = False
        self.allowResize = False
        self.allowMove = False
        self.tempBox = None
    
    
            
    def checkNoCollision( self ) :
        """
            Collision detection is done by verifying a linear (time) order of boxes while respecting canvas' boundaries
        """
        if self.allowDraw or self.allowResize :
            if self.tempX2 > self.gui.width - self.gui.canvasMargin :
                self.tempX2 = self.gui.width - self.gui.canvasMargin
            if self.tempX1 < self.gui.canvasMargin :
                self.tempX1 = self.gui.canvas.Margin
            if self.tempX2 < self.gui.canvasMargin:
                self.tempX2 = self.gui.canvasMargin
        # We only allow drawing from left to right for simplicity  
        if self.tempX2 < self.tempX1 and self.allowDraw :
            self.tempX2 = self.tempX1
        for i in range( len( self.gui.data.timeBoxes ) ) :
            if self.gui.data.timeBoxes[ i ].hidden :
                continue
            # If position is within an existing box return False
            if( self.gui.mode=="draw" and self.tempX1 >= self.gui.canvas.coords( self.gui.data.timeBoxes [i].getCanvasID() ) [0] and \
                self.tempX1 <= self.gui.canvas.coords( self.gui.data.timeBoxes [i].getCanvasID() ) [2] ) :
                return False
            # User draws rectangle from right towards left (not only draw mode here)
            if self.tempX2 <= self.tempX1  :
                if( self.gui.canvas.coords( self.gui.data.timeBoxes [i].getCanvasID() ) [2] >= self.tempX2 and \
                    self.gui.canvas.coords( self.gui.data.timeBoxes [i].getCanvasID() ) [2] <= self.tempX1 ) :
                    if( self.gui.mode=="move" and self.gui.data.timeBoxes [i].getCanvasID() != self.tempBox.getCanvasID() ):
                        return False
                    if( self.gui.mode!="move" and self.gui.data.timeBoxes [i].getCanvasID() != self.tempBox.getCanvasID() ) :
                        # +1 because we don't want overlapping rectangles
                        self.tempX2 = self.gui.canvas.coords( self.gui.data.timeBoxes [i].getCanvasID() ) [2] + 1
            # User draws rectangle from left towards the right
            if self.tempX2 >= self.tempX1  :
                if( self.gui.canvas.coords( self.gui.data.timeBoxes [i].getCanvasID() ) [0] <= self.tempX2 and \
                       self.gui.canvas.coords (self.gui.data.timeBoxes [i].getCanvasID() ) [0] >= self.tempX1) :
                    if( self.gui.mode=="move" and self.gui.data.timeBoxes [i].getCanvasID() != self.tempBox.getCanvasID() ) :
                        return False
                    if( self.gui.mode !="move" and self.gui.data.timeBoxes [i].getCanvasID() != self.tempBox.getCanvasID() ) :
                        # -1 because we don't want overlapping rectangles
                        self.tempX2 = self.gui.canvas.coords( self.gui.data.timeBoxes [i].getCanvasID() ) [0] - 1
        return True



    def checkCorner( self ) :
        """
            Check what box should be resized from which corner by min. euclid. dist. to corners
            Clicks must be placed less than 10 (eucl. dist.) from a corner to be valid
            corner scheme:
            (0,1)      (1,1)
            (0,0)      (1,0)
        """
        self.closestCorner = ( 0,0 )
        minDist = 10 # Legal and proper initialization (see above)
            
        for box in self.gui.data.timeBoxes :
            if not box.hidden :
                coords = self.gui.canvas.coords( box.getCanvasID() )

                lowerLeftDIST = sqrt( pow( self.tempX1-coords[ 0 ], 2 ) + pow( self.tempY1-coords[ 3 ], 2 ) )
                upperLeftDIST = sqrt( pow( self.tempX1-coords[ 0 ], 2 ) + pow( self.tempY1-coords[ 1 ], 2 ) )
                upperRightDIST =sqrt( pow( self.tempX1-coords[ 2 ], 2 ) + pow( self.tempY1-coords[ 1 ], 2 ) )
                lowerRightDIST =sqrt( pow( self.tempX1-coords[ 2 ], 2 ) + pow( self.tempY1-coords[ 3 ], 2 ) )
                
                if( lowerLeftDIST < minDist ) :
                    minDist = lowerLeftDIST
                    self.tempBox = box
                    #self.closestCorner = (0,0)
                    self.tempX1 = coords[ 0 ]
                    self.tempY1 = coords[ 3 ]
                    self.tempX2=self.tempX1
                    self.tempY2=self.tempY1
                    self.allowResize = True
                    self.currentCorner="lowerLeft"

                if( upperLeftDIST < minDist ) :
                    minDist = upperLeftDIST
                    self.tempBox = box
                    #self.closestCorner = (0,1)
                    self.tempX1 = coords[ 0 ]
                    self.tempY1 = coords[ 1 ]
                    self.tempX2=self.tempX1
                    self.tempY2=self.tempY1
                    self.allowResize = True
                    self.currentCorner="upperLeft"

                if( upperRightDIST < minDist ) :
                    minDist = upperRightDIST
                    self.tempBox = box
                    #self.closestCorner = (1,1)
                    self.tempX1 = coords[ 2 ]
                    self.tempY1 = coords[ 1 ]
                    self.tempX2=self.tempX1
                    self.tempY2=self.tempY1
                    self.allowResize = True
                    self.currentCorner="upperRight"

                if( lowerRightDIST < minDist ) :
                    minDist = lowerRightDIST
                    self.tempBox = box
                    #self.closestCorner = (1,0)
                    self.tempX1 = coords[ 2 ]
                    self.tempY1 = coords[ 3 ]
                    self.tempX2=self.tempX1
                    self.tempY2=self.tempY1
                    self.allowResize = True
                    self.currentCorner="lowerRight"
        return


    
    def roundToHalf( self, number ) :
        """
            rounds any given float to the closest float, that ends with .5 and returns at least 0.5
        """
        rounded = round(number)
        # If rounded <= 0 :
        #    return 0.5
        if rounded < number :
            rounded += 0.5
        else :
            rounded += -0.5
        return rounded


    
    def mouseDown( self, event ) :
        inverseScaleFunction = self.gui.renderer.inverseScaleFunction
        self.tempX1 = self.tempX2 = self.gui.canvas.canvasx( event.x ) 
        self.tempY1 = self.tempY2 = self.gui.canvas.canvasy( event.y )
        """
            check:
                (a) user chose 'draw' thru GUI 
                (b) the initial corner was not placed within an existing rectangle
                (c) the initial corner of the rectangle was not placed self.gui.canvasMargin to the sides of the canvas
            if all conditions (a) - (c) match:
                    allow drawing
        """
        if self.gui.mode =="draw" and not \
            ( ( self.tempX1 < self.gui.canvasMargin ) or \
              ( self.tempX1 > self.gui.width - self.gui.canvasMargin ) ) :
            noCollision = self.checkNoCollision()
            correctSpacing = True
            requestedCoordinate = inverseScaleFunction( self.tempX1, 0 )[ 0 ]
            for i in range( len( self.gui.data.timeBoxes ) ) :
                if self.gui.data.timeBoxes[ i ].hidden :
                    continue
                # A new box must start at least a unit length away from any existing box
                x1 =self.gui.canvas.coords( self.gui.data.timeBoxes[ i ].getCanvasID() )[ 0 ]
                x2 =self.gui.canvas.coords( self.gui.data.timeBoxes[ i ].getCanvasID() )[ 2 ]
                x1= inverseScaleFunction( x1, 0 )[ 0 ]
                x2= inverseScaleFunction( x2, 0 )[ 0 ]
                if requestedCoordinate > x2 and requestedCoordinate - x2 < 1.0 :
                    correctSpacing=False
                if requestedCoordinate < x1 and x1 - requestedCoordinate < 1.0 :
                    correctSpacing=False
            if noCollision and correctSpacing :
                self.allowDraw = True
                
        if self.gui.mode == "resize" :
            self.checkCorner()
            
        if self.gui.mode == "move" :
            for box in self.gui.data.timeBoxes :
                if not box.hidden :
                    coords = self.gui.canvas.coords( box.getCanvasID() )
                    if( coords[ 0 ] <= self.tempX1 <=coords[ 2 ] and \
                        coords[ 1 ] <= self.tempY1 <= coords[ 3 ] ) :
                        self.tempBox = box
                        self.allowMove = True
                        
        
            
    def mouseMotion( self, event ) :
        self.tempX2 = self.gui.canvas.canvasx( event.x )
        self.tempY2 = self.gui.canvas.canvasy( event.y )
        inverseScaleFunction = self.gui.renderer.inverseScaleFunction
        scaleFunction = self.gui.renderer.scaleFunction
            
        if self.allowDraw :
            if self.tempBox != None :
                self.gui.renderer.deleteWidget( self.tempBox.getCanvasID() )
            # We only permit boxes that start (on the left) on exactly a .5 time coordinate, this code forces this constraint
            dataCoords = inverseScaleFunction( self.tempX1,  self.tempY1 )
            roundedDataCoords = ( self.roundToHalf( dataCoords[ 0 ] ), dataCoords[ 1 ] )
            if roundedDataCoords[ 0 ] < 0.5 :
                roundedDataCoords[ 0 ] = 0.5
            roundedScreenCoords = ( scaleFunction( roundedDataCoords[ 0 ],0 )[ 0 ], scaleFunction( 0, roundedDataCoords[ 1 ] ) [ 1 ] )
            self.tempX1= roundedScreenCoords[ 0 ]
            self.tempY1 = roundedScreenCoords[ 1 ]
            self.checkNoCollision()
            
            timeBoxCanvasID = self.gui.renderer.drawTimeBox( self.tempX1, self.tempY1, self.tempX2, self.tempY2 )
            x0 = inverseScaleFunction( self.tempX1, 0 )[ 0 ]
            x1 = inverseScaleFunction( self.tempX2, 0 )[ 0 ]
            y0 = inverseScaleFunction( 0, self.tempY1 )[ 1 ]
            y1 = inverseScaleFunction( 0, self.tempY2 )[ 1 ]
            self.tempBox = TimeBox(   False, \
                                timeBoxCanvasID, \
                                ( y0+y1 )/2.0, \
                                abs((y1 - ( y0+y1 )/2.0 )/2.0), \
                                abs( x1-x0 ), \
                                x0 )
                                
        if self.allowResize :
            self.checkNoCollision()
            oldCoordinates = self.gui.canvas.coords( self.tempBox.getCanvasID() )
            newCoordinates = oldCoordinates
            # Compute coordinates of the resized box
            if self.currentCorner =="upperLeft" :
                tempX2Data = inverseScaleFunction( self.tempX2, 0 )[ 0 ]
                # Guarantee, that resizing a left hand side handle will move it to a x.5 x-coordinate
                if tempX2Data % 0.5 != 0.0 :
                        roundedX2Data = self.roundToHalf( tempX2Data )
                        roundedX2Screen = scaleFunction( roundedX2Data, 0 )[ 0 ]
                        newCoordinates = [ roundedX2Screen, self.tempY2, oldCoordinates[ 2 ], oldCoordinates[ 3 ] ]
                else :
                    newCoordinates = [ self.tempX2, self.tempY2, oldCoordinates[ 2 ], oldCoordinates[ 3 ] ]
            if self.currentCorner =="lowerRight" :
                newCoordinates = [ oldCoordinates[ 0 ], oldCoordinates[ 1 ], self.tempX2, self.tempY2 ]
            if self.currentCorner =="lowerLeft" :
                tempX2Data = inverseScaleFunction( self.tempX2, 0 )[ 0 ]
                # Guarantee, that resizing a left hand side handle will move it to a x.5 x-coordinate
                if tempX2Data % 0.5 != 0.0 :
                    roundedX2Data = self.roundToHalf( tempX2Data )
                    roundedX2Screen = scaleFunction( roundedX2Data, 0 )[ 0 ]
                    newCoordinates = [ roundedX2Screen, oldCoordinates[ 1 ], oldCoordinates[ 2 ], self.tempY2 ]
                else :
                    newCoordinates = [ self.tempX2, oldCoordinates[ 1 ], oldCoordinates[ 2 ], self.tempY2 ]
            if self.currentCorner =="upperRight" :
                newCoordinates = [ oldCoordinates[ 0 ], self.tempY2, self.tempX2, oldCoordinates[ 3 ] ]
            dataCoords1 = list( inverseScaleFunction( newCoordinates[ 0 ], newCoordinates[ 1 ] ) )
            dataCoords2 = list( inverseScaleFunction( newCoordinates[ 2 ], newCoordinates[ 3 ] ) )
            # Guarantee .5 distance to the origin on x-axis
            if dataCoords1[ 0 ] < 0.5 :
                dataCoords1[ 0 ] = 0.5
                newCoordinates[ 0 ] = scaleFunction( dataCoords1[ 0 ], 0 )[ 0 ]
            if dataCoords2[ 0 ] < 0.5 :
                dataCoords2[ 0 ] = 0.5
                newCoordinates[ 2 ] = scaleFunction( dataCoords2[ 0 ], 0 )[ 0 ]
            # Guarantee last time point + 0.5 as a right border
            if dataCoords1[ 0 ] > self.gui.renderer.timepoints - 1 + 0.5 :
                dataCoords1[ 0 ] = self.gui.renderer.timepoints - 1 + 0.5
                newCoordinates[ 0 ] = scaleFunction( dataCoords1[ 0 ], 0 )[ 0 ]
            if dataCoords2[ 0 ] > self.gui.renderer.timepoints - 1 + 0.5 :
                dataCoords2[ 0 ] = self.gui.renderer.timepoints - 1 + 0.5
                newCoordinates[ 2 ] = scaleFunction( dataCoords2[ 0 ], 0 )[ 0 ]
            # Guarantee, that the resized box does not violate the inter box spacing constraint
            boxSpacing = 1.0
            for i in range( len( self.gui.data.timeBoxes ) ) :
                if self.gui.data.timeBoxes[ i ].hidden :
                    continue
                if self.tempBox.getCanvasID() != self.gui.data.timeBoxes[ i ].getCanvasID()  :
                    x0 = inverseScaleFunction( self.gui.canvas.coords( self.gui.data.timeBoxes[ i ].getCanvasID() ) [ 0 ], 0 )[ 0 ]
                    x1 = inverseScaleFunction( self.gui.canvas.coords( self.gui.data.timeBoxes[ i ].getCanvasID() ) [ 2 ], 0 )[ 0 ]
                    if( dataCoords2[ 0 ] < x0 and x0 - dataCoords2[ 0 ] < boxSpacing ) :
                        dataCoords2 = ( x0 - boxSpacing, dataCoords2[ 1 ] )
                        newCoordinates[ 2 ] = scaleFunction(dataCoords2[ 0 ], dataCoords2[ 1 ])[ 0 ]
                    if dataCoords1[ 0 ] > x1 and dataCoords1[ 0 ] -x1 < boxSpacing :     
                        dataCoords1 = ( x1 + boxSpacing, dataCoords1[ 1 ] )
                        newCoordinates[ 0 ] = scaleFunction( dataCoords1[ 0 ], 0 )[ 0 ]
            boxWidth= dataCoords2[ 0 ] - dataCoords1[ 0 ]
            if boxWidth < 1.0 :
                newCoordinates[ 0 ] = oldCoordinates[ 0 ]
                newCoordinates[ 2 ] = oldCoordinates[ 2 ]

            # Draw the resized box
            self.gui.canvas.coords( self.tempBox.getCanvasID(), newCoordinates[ 0 ], newCoordinates[ 1 ],newCoordinates[ 2 ], newCoordinates[ 3 ] )
            # Set the parameters in the box model
            x0 = inverseScaleFunction( newCoordinates[ 0 ], 0 )[ 0 ]
            x1 = inverseScaleFunction( newCoordinates[ 2 ], 0 )[ 0 ]
            y0 = inverseScaleFunction( 0, newCoordinates[ 1 ] )[ 1 ]
            y1 = inverseScaleFunction( 0, newCoordinates[ 3 ] )[ 1 ]
            self.tempBox.mean = ( y0+y1 )/2.0
            self.tempBox.variance = abs( ( y1 - ( y0+y1 )/2.0 )/2.0 )
            self.tempBox.duration = abs( x1-x0 )
            self.tempBox.begin = x0
            # Update the positions of the resize handles and update their data model
            boxCoords = self.gui.canvas.coords( self.tempBox.getCanvasID() )
            for box in self.gui.data.timeBoxes :
                if not box.hidden :
                    for handle in box.handles :
                        if( handle.associatedTimeBoxID == self.tempBox.getCanvasID() ) :
                            if( handle.position == "downLeft" ) :
                                self.gui.renderer.move( handle.canvasID, boxCoords[ 0 ]-3, boxCoords[ 3 ]-3, boxCoords[ 0 ]+3, boxCoords[ 3 ]+3 )
                                handle.x=inverseScaleFunction( boxCoords[ 0 ], 0 )[ 0 ]
                                handle.y=inverseScaleFunction( 0, boxCoords[3] )[ 1 ]
                            if( handle.position == "upLeft" ) :
                                self.gui.renderer.move( handle.canvasID, boxCoords[ 0 ]-3, boxCoords[ 1 ]-3, boxCoords[ 0 ]+3, boxCoords[ 1 ]+3 )
                                handle.x=inverseScaleFunction( boxCoords[ 0 ], 0 )[ 0 ]
                                handle.y=inverseScaleFunction( 0, boxCoords[ 1 ] )[ 1 ]
                            if( handle.position == "upRight" ) :
                                self.gui.renderer.move( handle.canvasID, boxCoords[ 2 ]-3, boxCoords[ 1 ]-3, boxCoords[ 2 ]+3, boxCoords[ 1 ]+3 )
                                handle.x=inverseScaleFunction( boxCoords[ 2 ], 0 )[ 0 ]
                                handle.y=inverseScaleFunction( 0, boxCoords[ 1 ] )[ 1 ]
                            if( handle.position == "downRight" ) :
                                self.gui.renderer.move( handle.canvasID, boxCoords[ 2 ]-3, boxCoords[ 3 ]-3, boxCoords[ 2 ]+3, boxCoords[ 3 ]+3 )
                                handle.x=inverseScaleFunction( boxCoords[ 2 ], 0 )[ 0 ]
                                handle.y=inverseScaleFunction( 0, boxCoords[ 3 ] )[ 1 ]
        
        if self.allowMove :
            xTranslation = round( inverseScaleFunction( self.tempX2, 0 )[ 0 ] - inverseScaleFunction( self.tempX1, 0 )[ 0 ] )
            yTranslation = inverseScaleFunction( 0, self.tempY2 )[ 1 ] - inverseScaleFunction( 0, self.tempY1 )[ 1 ]
            # Boxes left side must leave at least .5 distance to zero
            if self.tempBox.begin + xTranslation < 0.5 :
                xTranslation = self.roundToHalf( - self.tempBox.begin+0.5 )
                xTranslation = - self.tempBox.begin+0.5            
            # Boxes right side must end at most .5 past the last time point.
            if self.tempBox.begin + self.tempBox.duration + xTranslation > self.gui.renderer.timepoints + 0.5 :
                xTranslation = ( self.gui.renderer.timepoints ) - ( self.tempBox.begin + self.tempBox.duration )
            # The translation is computed as coordinate difference - position of origin
            xTranslationScreenCoords = scaleFunction( xTranslation , 0 )[ 0 ] - scaleFunction( 0, 0 )[ 0 ]
            yTranslationScreenCoords = scaleFunction( 0, yTranslation )[ 1 ] - scaleFunction( 0, 0 )[ 1 ]
            move=True
            # Guarantee non-overlapping boxes
            for box in self.gui.data.timeBoxes :
                if box.hidden :
                    continue
                if box != self.tempBox :
                    # Moved box would start within another box
                    if self.tempBox.begin + xTranslation > box.begin and \
                       self.tempBox.begin + xTranslation < box.begin+box.duration :
                        move = False
                    if self.tempBox.begin + xTranslation == box.begin :
                        move = False
                    # Moved box would end in another box
                    if self.tempBox.begin + self.tempBox.duration + xTranslation > box.begin and \
                       self.tempBox.begin + self.tempBox.duration + xTranslation < box.begin+box.duration :
                        move = False
                    if self.tempBox.begin + self.tempBox.duration + xTranslation == box.begin + box.duration :
                        move = False
                    # After move another box would be contained in the moved box
                    if box.begin > self.tempBox.begin + xTranslation and \
                        box.begin + box.duration < self.tempBox.begin + self.tempBox.duration + xTranslation :
                            move = False
                    
            # Guarantee 1.0 box spacing constraint
            boxSpacing = 1.0
            for box in self.gui.data.timeBoxes :
                if box.hidden :
                    continue
                if box != self.tempBox :
                    # moved box is left of another box
                    if box.begin > self.tempBox.begin + self.tempBox.duration + xTranslation and \
                        box.begin - (self.tempBox.begin + self.tempBox.duration + xTranslation) < 1.0 :
                            move = False
                    # moved box is right of another box
                    if box.begin + box.duration < self.tempBox.begin + xTranslation and \
                        self.tempBox.begin + xTranslation - ( box.begin + box.duration ) < 1.0 :
                            move = False
            # Move the box actually
            if( move ) :
                x0 = scaleFunction( self.tempBox.begin, 0 )[ 0 ]
                y0 = scaleFunction( 0, self.tempBox.mean-2*self.tempBox.variance )[ 1 ]
                x1 = scaleFunction( self.tempBox.begin, 0 )[ 0 ] + scaleFunction( self.tempBox.duration, 0 )[ 0 ]-self.gui.canvasMargin
                y1 = scaleFunction( 0, self.tempBox.mean+2*self.tempBox.variance )[ 1 ]
                self.gui.canvas.coords( self.tempBox.getCanvasID(), x0+xTranslationScreenCoords, y0+yTranslationScreenCoords, \
                                                                    x1+xTranslationScreenCoords, y1+yTranslationScreenCoords )


    
    def mouseUp( self, event ) :
        """
            At the mouse up event changes really manifest in the data model (and persist there and on the canvas).
        """
        inverseScaleFunction = self.gui.renderer.inverseScaleFunction
        dataCoords1 = inverseScaleFunction( self.tempX1, self.tempY1 )
        dataCoords2 = inverseScaleFunction( self.tempX2, self.tempY2 )
        boxWidth= dataCoords2[ 0 ] - dataCoords1[ 0 ]
        
        if self.allowDraw and boxWidth >= 1 :
            inverseScaleFunction = self.gui.renderer.inverseScaleFunction
            timeBoxCanvasID = self.gui.renderer.drawTimeBox( self.tempX1, self.tempY1, self.tempX2, self.tempY2 )
        
            x0 = inverseScaleFunction( self.tempX1, 0 )[ 0 ]
            x1 = inverseScaleFunction( self.tempX2, 0 )[ 0 ]
            y0 = inverseScaleFunction( 0, self.tempY1 )[ 1 ]
            y1 = inverseScaleFunction( 0, self.tempY2 )[ 1 ]
            newBox = TimeBox(   False, \
                                timeBoxCanvasID, \
                                ( y0+y1 )/2.0, \
                                abs( ( y1 - ( y0+y1 )/2.0 )/2.0 ), \
                                abs( x1-x0 ), \
                                x0 )
            
            self.gui.data.addBox( newBox )
            self.allowDraw = False
            self.tempX1 = self.tempY1 = self.tempX2 = self.tempY2 = 0
            self.gui.renderer.deleteWidget( self.tempBox.getCanvasID() )
            self.tempBox = None
            self.queryChanged()
        
        if self.allowDraw and boxWidth < 1 and self.tempBox != None :
            self.gui.renderer.deleteWidget( self.tempBox.getCanvasID() )
            self.allowDraw=False
        
        if self.allowResize :
            self.allowResize = False
            self.tempBox = None
            self.queryChanged()
        
        if self.allowMove :
            coords = self.gui.canvas.coords( self.tempBox.getCanvasID() )
            x0 = inverseScaleFunction( coords[ 0 ], 0 )[ 0 ]
            x1 = inverseScaleFunction( coords[ 2 ], 0 )[ 0 ]
            y0 = inverseScaleFunction( 0, coords[ 1 ] )[ 1 ]
            y1 = inverseScaleFunction( 0, coords[ 3 ] )[ 1 ]
            self.tempBox.mean = ( y0+y1 ) / 2.0
            self.tempBox.variance = abs( ( y1 - ( y0+y1 )/2.0 )/2.0 )
            self.tempBox.duration = abs( x1-x0 )
            self.tempBox.begin = x0
            self.allowMove = False
            self.gui.data.recomputeHiddenBoxes()
            self.queryChanged()


    def queryChanged( self ) :
        self.gui.renderer.hide = []
        self.gui.renderer.show = copy.copy( self.gui.renderer.timecourses )
        self.gui.data.queryResult = copy.copy( self.gui.data.timeCourses )
    
        # reproduction of basic deterministic queries from Hochheiser
        if self.gui.deterministicQuery :
            show = [] #update for the show list
            hide = [] #update for the hide list
            #check which currently shown time courses need to be hidden according to query
            for t in range( 0, len( self.gui.renderer.show ) ) :
                tc  = self.gui.renderer.show[ t ]
                showTC = True
                
                for time in range( 0, len( tc ) ) :
                    for box in self.gui.data.timeBoxes :
                        if box.hidden :
                            continue#with next box
                        else: 
                            if time+1 >= box.begin and time+1 <= box.begin+box.duration :
                                if float( tc[ time ] ) >= box.mean - 2*box.variance and \
                                   float( tc[ time ] ) <= box.mean + 2*box.variance :
                                    continue#with next box
                                else :
                                    showTC = False
                                    break
                if not showTC :
                    hide.append( t )
            #check which currently hidden time courses need to be shown according to query            
            for t in range( 0, len( self.gui.renderer.hide ) -1 ) :
                tc  = self.gui.renderer.hide[ t ]
                hideTC = True
                for time in range( 0, len( tc ) -1 ) :
                    for box in self.gui.data.timeBoxes :
                        if box.hidden :
                            continue
                        else: 
                            if time+1 >= box.begin and time+1 <= box.begin+box.duration :
                                if float( tc[ time ] ) >= box.mean - 2*box.variance and \
                                   float( tc[ time ] ) <= box.mean + 2*box.variance :
                                    continue
                                else :
                                    hideTC = False
                                    break
                if not hideTC :
                    show.append( t )
            
            self.gui.renderer.update( show, hide )
            # do not execute the probabilistic query
            return
        
        # else: proceed with probabilistic query
        self.gui.query = GQLQuery()
        densityTypes = []
        for box in self.gui.data.timeBoxes :
            if box.hidden :
                densityTypes.append ( ghmmwrapper.uniform )
                continue
            else :
                densityTypes.append( ghmmwrapper.normal )
                continue
        timecourseData = []
        for i in range( len( self.gui.renderer.timecourses ) ) :
            timecourseData.append( map( float, self.gui.renderer.timecourses[ i ] ) )
        self.gui.query.addListener( self.gui.renderer )
        sequences = GQL.ProfileSet()
        sequences.ReadDataFromCaged( self.gui.dataFile )
        self.gui.query.newQuery( len( self.gui.data.timeBoxes ), self.gui.data.getMeanVector(), 
            self.gui.data.getVarianceVector(), self.gui.data.getDurationsVector(), False, densityTypes )
        self.gui.query.setData( sequences )
        self.gui.query.newQuery( len( self.gui.data.timeBoxes ), self.gui.data.getMeanVector(), 
            self.gui.data.getVarianceVector(), self.gui.data.getDurationsVector(), False, densityTypes )
        self.gui.query.setRankCutoff( self.gui.rankSlider.get() )

#------------------------------------------------------------------------------------------------------
#******************************************************************************************************
#------------------------------------------------------------------------------------------------------



#------------------------------------------------------------------------------------------------------
#******************************************************************************************************
#------------------------------------------------------------------------------------------------------
class Rendering :
    
    def __init__( self, gui ) :
        self.xLabels=self.yLabels = []
        self.xAxis=self.yAxis=None
        self.gui = gui
        self.canvas = gui.canvas
        self.timecourses = []            #all timecourses
        self.highestValue = 1.0
        self.lowestValue = 0.0
        self.timepoints = 1
        self.scaleFunction = self.linearScale
        self.inverseScaleFunction=self.inverseLinearScale
        self.show = [] # query result
        self.hide = [] # query result
    
    
    
    def clearCanvas( self ) :
        """
            Clear the canvas and forget its objects
        """
        self.canvas.delete( "all" )
        self.timecourses = []
        self.yLabels=self.xLabels = []
        self.canvas.update_idletasks()
    
    
    
    def update( self, showProfiles, hideProfiles ) :
        """
            Interface implementation for interaction with GQLQuery
        """
        for tc in hideProfiles :
            self.hide.append( self.timecourses[ tc ] )
            self.show.remove( self.timecourses[ tc ] )
            self.gui.data.queryResult.remove( self.gui.data.timeCourses[ tc ] )
        for tc in showProfiles :
            self.gui.data.queryResult.append( self.gui.data.timeCourses[ tc ] )
            self.show.append( self.timecourses[ tc ] )
            self.hide.remove( self.timecourses[ tc ] )
        # redraw the canvas with the query results
        self.canvas.delete( "timecourse" )
        self.drawTimeCourseData( self.scaleFunction, self.hide, 'gray75' )
        self.drawTimeCourseData( self.scaleFunction, self.show, 'red' )
        self.adjustRenderingOrder()
    
    
    
    def findMinMax( self ) :
        """
            Find the minimum and maximum values and number of time points
        """
        self.timepoints = len( self.timecourses[ 0 ] )+1 # assuming the length of all time courses has been equalized
        for i in xrange( len( self.timecourses ) ) :
            for j in self.timecourses[ i ] :
                if( j != '' ) : # ignore missing values for scaling
                    if self.highestValue < float( j )  :
                        self.highestValue = float( j )
                    if self.lowestValue > float( j ) and abs( float( j ) ) < 9999 :
                        self.lowestValue= float( j )
                        

    
    def maxScale( self, x, y ) :
        """
            Dynamic scaling functions that map the range of the data space to available display coordinate space 
            The method returns a coordinate tuple.
        """
        x = self.gui.canvasMargin+x*( ( self.gui.width-2*self.gui.canvasMargin ) / float( self.timepoints ) ) # normalized by number of time points
        y = self.gui.height / 2-y*( ( ( self.gui.height / 2 )-2*self.gui.canvasMargin ) / ceil( self.highestValue ) )  # normalized by highest value
        return (x, y)
    
    
    
    def linearScale ( self, x, y ) :
        """
            The method returns a coordinate tuple.
        """
        horizontal = ( self.gui.width - 2 * self.gui.canvasMargin ) / float( self.timepoints )
        vertical = ( self.gui.height - 2 * self.gui.canvasMargin ) / ( max( abs( floor( self.lowestValue ) ), ceil( self.highestValue ) ) * 2.0 )
        linScaleFactor = min( vertical, horizontal )
        x = self.gui.canvasMargin + x * linScaleFactor
        y = self.gui.height/2.0 - y * linScaleFactor        
        return ( x, y )
    
    
    
    def inverseMaxScale( self, x, y ) :
        """ 
            Dynamic scaling functions, that map the display coordinate space to the data space 
            The method returns a coordinate tuple.
        """
        x = ( x-self.gui.canvasMargin ) * float( self.timepoints ) / ( self.gui.width - 2 * self.gui.canvasMargin )
        y = -( y-self.gui.height/2 ) / ( ( ( self.gui.height/2 ) - 2 * self.gui.canvasMargin ) / float( ceil( self.highestValue ) ) )
        return ( x, y )
    
    
    
    def inverseLinearScale( self, x, y ) :
        """
            Dynamic scaling functions, that map the display coordinate space to the data space 
            The method returns a coordinate tuple.
        """
        horizontal = ( self.gui.width - 2 * self.gui.canvasMargin ) / float( self.timepoints )
        vertical = ( self.gui.height - 2 * self.gui.canvasMargin ) / ( max( abs( floor( self.lowestValue ) ), ceil( self.highestValue ) ) * 2.0 )
        linScaleFactor = min( vertical, horizontal )
        
        x = - ( self.gui.canvasMargin - x ) * 1 / float( linScaleFactor )
        y = - ( y - ( self.gui.height / 2.0 ) ) / float( linScaleFactor )
        return ( x, y ) 
    
    
    
    def adjustRenderingOrder( self ) :
        """
            This function implements a user choice to display the time boxes above or below the 
            time courses. It is also called from various functions, that render graphical elements.
        """
        self.canvas.tag_lower( "caption", "handle" )
        self.canvas.tag_lower( "axis", "caption" )
        self.canvas.tag_lower( "gridline", "axis" )
        self.canvas.tag_lower( "timecourse", "gridline" )
        if self.gui.boxLevelVar == "back" :
            self.canvas.tag_lower( "timebox", "timecourse" )
        if self.gui.boxLevelVar == "front" :
            self.canvas.tag_lower( "timebox", "timecourse" )
            self.canvas.tag_lower( "timecourse", "timebox" )
                
    
        
    def redisplay( self ) :
        """
            This function is issued from the GUI when
                (a) The visualization scale has been changed
                (b) The display style of the time courses has been changed
            In any case this function handles the redrawing of the canvas.
        """                
        if self.gui.data.timeCourses == [] :
            return
        if self.gui.viewScaleVar.get() == "linear" :
                self.scaleFunction = self.linearScale
                self.inverseScaleFunction=self.inverseLinearScale
        if self.gui.viewScaleVar.get() == "max" :
                self.scaleFunction = self.maxScale
                self.inverseScaleFunction=self.inverseMaxScale
        
        canvasObjects = self.canvas.find_all()
        for object in canvasObjects :
            delete=True
            for box in self.gui.data.timeBoxes :
                if not box.hidden :
                    for handle in box.handles :
                        if handle.canvasID == object :
                            delete=False
                            break
            if delete :
                self.canvas.delete( object )
        
        # first two columns only contain non-display information
        self.timecourses = []
        for i in range( len( self.gui.data.timeCourses ) ) :
            #result = string.split (self.gui.data.timeCourses[i]) # split values separated by tabs or white space
            result =self.gui.data.timeCourses[ i ].split( '\t' ) # split values separated by tabs or white space
            self.timecourses.append( result[ 2: ] )
        
        # Now display (either the query result or the data set)
        if self.show != [] or self.hide != [] :
            self.drawTimeCourseData( self.scaleFunction, self.hide, 'gray75' )
            self.drawTimeCourseData( self.scaleFunction, self.show, 'red' )
        else:
            self.drawTimeCourseData( self.scaleFunction, self.timecourses )
        self.drawAxes( self.scaleFunction )
        
        for tb in self.gui.data.timeBoxes :
            if not tb.hidden :
                x0=self.scaleFunction( tb.getBegin(), 0 )[ 0 ]
                y0=self.scaleFunction( 0, tb.getMean()- 2*tb.getVariance() )[ 1 ]
                x1=self.scaleFunction( tb.getBegin()+tb.getDuration(), 0 )[ 0 ]
                y1=self.scaleFunction( 0, tb.getMean() + 2*tb.getVariance() )[ 1 ]
                tb.setCanvasID( self.gui.renderer.drawTimeBox( x0, y0, x1, y1 ) )
                for handle in tb.handles :
                    x=self.scaleFunction( handle.x, 0 )[ 0 ]
                    y=self.scaleFunction( 0, handle.y )[ 1 ]
                    self.gui.renderer.move( handle.canvasID, x-3, y-3, x+3, y+3 )
                    handle.associatedTimeBoxID = tb.getCanvasID()
        # readjust rendering order on the canvas
        self.adjustRenderingOrder()                    
    
                                    
        
    def display( self ) :
        """
            The essential difference to the redisplay function, is
            that 'display' deletes all canvas objects prior to building
            all canvas objects and drawing everything from scratch.
        """
        self.clearCanvas()
        if self.gui.data.timeCourses == [] :
            return
        self.highestValue = 1.0
        self.lowestValue = 0.0
        # first two columns only contain non-display information
        self.timecourses = []
        for i in range( len( self.gui.data.timeCourses ) ) :
            #result = string.split (self.gui.data.timeCourses[i]) # split values separated by tabs or white space
            result =self.gui.data.timeCourses[ i ].split( '\t' ) # split values separated by tabs or white space
            self.timecourses.append( result[ 2: ] )
        self.show = copy.copy( self.timecourses ) # initialize the query result, no query = all data
        # now display
        f = None
        if self.gui.viewScaleVar.get() == "linear" :
            f = self.linearScale
        if self.gui.viewScaleVar.get() == "max" :
            f = self.maxScale
        self.drawTimeCourseData( f, self.timecourses )
        self.drawAxes( f )
        
    
    
    def drawTimeBox( self, x0, y0, x1, y1 ) :
        """ 
            Expects four coordinates (x0,y0),(x1,y1) in this order x0,y0,x1,y1 as arguments
            then draws a box on the canvas and returns its canvas ID
        """
        canvasID = self.gui.canvas.create_rectangle( x0, y0, x1, y1, outline="royalblue", fill="royalblue", tag="timebox" )
        self.adjustRenderingOrder()
        return canvasID

    
    
    def deleteWidget( self, timeBoxID ) :
        self.gui.canvas.delete( timeBoxID )

        

    def setOutline( self, canvasItemID, thickness ) :
        """
            Manipulates the thickness of time box outlines
        """
        self.gui.canvas.itemconfigure( canvasItemID, width=thickness )
        
    
            
    def move( self, canvasObjectID, x0, y0, x1, y1 ) :
        """
            Moves the specified canvas item to the given coordinates
        """
        self.gui.canvas.coords( canvasObjectID, x0, y0, x1, y1 )
    
    
    
    def drawTimeCourseData( self, scaleMethod, timecourses, color='red' ) :
        """
            This method draws all the time courses using a defined scaling method
            it expects a list of time courses (list of list of numerical values) format
            and a string with a tk compatible color.
            Currently the function is used in two use cases: 1) drawing initial (unqueried data)
            and 2) drawing queried data (using two colors)
        """
        self.findMinMax()
        
        for timeCourse in timecourses :
            time = 1
            coordinatesDrawList = []
            coordinates = ()
            
            # scale the coordinates list for drawing
            for val in timeCourse :
                if val != '' and float( val ) != -9999.99 :
                    coords = scaleMethod( time, float( val ) )
                    coordinatesDrawList.append( ( coords[ 0 ], coords[ 1 ] ) )
                else :
                    missingData = -9999.99
                    coordinatesDrawList.append( missingData )
                time = time+1
            
            # actually draw the coordinates list
            for t in xrange( time-1 ) :
                if coordinatesDrawList[ t ] != -9999.99 :
                    coordinates += coordinatesDrawList[ t ]
            self.canvas.tk.call( self.canvas._w, 'create', 'line', coordinates, '-width', 0 ,'-fill', color, \
                                    '-smooth', self.gui.interpolationVar.get(), '-joinstyle', 'round', '-tag', 'timecourse' )
        # adjust rendering order
        self.adjustRenderingOrder()
        

            
    def drawAxes( self, scaleMethod ) :
        """
            Draw marks, captions and the axes themselves
        """
        # create axes
        self.xAxis = self.canvas.create_line(  self.gui.canvasMargin, self.gui.height / 2, self.gui.width-self.gui.canvasMargin+10, self.gui.height / 2, \
                                                arrow="last", width=2, fill="gray50", tags="axis" )
        self.yAxis = self.canvas.create_line(  self.gui.canvasMargin, self.gui.canvasMargin, self.gui.canvasMargin, \
                                                self.gui.height-self.gui.canvasMargin, arrow="first", width=2, fill="gray50", tags="axis" )
    
        # create *value marks* and *captions* on the *x-axis* and draw the vertical grid lines
        for i in xrange( self.timepoints ) :
            # draw vertical grid lines
            if( i > 0 ) :
                self.canvas.create_line(  scaleMethod( i, 0 )[ 0 ], scaleMethod( 0, int( floor( self.lowestValue ) ) )[ 1 ], scaleMethod( i, 0 )[ 0 ], \
                scaleMethod( 0, int( ceil( self.highestValue ) ) )[ 1 ], width=0, fill="gray75", tags="gridline" )
            # create scale *caption* on x-axis
            if i % 5 != 0 and i > 0 :
                self.xLabels.append( self.canvas.create_text(  scaleMethod( i, 0 )[ 0 ], self.gui.height / 2, \
                                                                text=i, font="Helvetica 16 normal", tags="caption" ) )
            # create every 5 time steps a bold number for orientation
            if i % 5 == 0  :
                self.xLabels.append( self.canvas.create_text( scaleMethod( i, 0 )[ 0 ], self.gui.height / 2, \
                                                                text=i, font="Helvetica 16 bold", tags="caption" ) )
        
        # create value *marks* on the *y-axis* and draw the horizontal grid lines
        for i in xrange( int( floor( self.lowestValue ) ), int( ceil( self.highestValue ) ) +1 ) :
            if i != 0:
                self.canvas.create_line( self.gui.canvasMargin, scaleMethod( 0,i )[ 1 ], scaleMethod( self.timepoints-1, 0 )[ 0 ], \
                                         scaleMethod( 0,i )[ 1 ], width=0, fill="gray75", tags="gridline" )
            if i % 5 != 0 :
                self.yLabels.append( self.canvas.create_text(  self.gui.canvasMargin, scaleMethod( 0,i )[ 1 ], text=float( i ), \
                                                                font="Helvetica 16 normal", tags="caption") )
            if i % 5 == 0 and i != 0:
                self.yLabels.append( self.canvas.create_text(  self.gui.canvasMargin, scaleMethod( 0,i )[ 1 ], text=float( i ), \
                                                                font="Helvetica 16  bold", tags="caption" ) )
#------------------------------------------------------------------------------------------------------
#******************************************************************************************************
#------------------------------------------------------------------------------------------------------
        


#------------------------------------------------------------------------------------------------------
#******************************************************************************************************
#------------------------------------------------------------------------------------------------------
class GUIMaster( Frame ) :

    def __init__( self, master=None ) :        
        Frame.__init__( self, master )
        self.scriptLocation = self.getScriptLocation()
        self.master.title( "pGQL query tool" )
        self.dataFile=""
        self.width=800
        self.height=600 
        self.mode = "no mode"
        self.data = Data( self )
        self.query = GQLQuery()
        self.deterministicQuery = False # default are probabilistic queries
        self.interaction = Interaction( self )
        self.canvas = Canvas( self.master, width=self.width, height=self.height, background="white", borderwidth=4, relief=GROOVE )
        self.canvas.grid( row=2, column=1, rowspan=2, sticky='N'+'E'+'S'+'W' )
        self.canvasMargin = 20 # constant margin of the canvas
        self.renderer = Rendering( self )
        self.createGUI()

        # event handling
        Widget.bind( self.canvas, "<Button-1>", self.interaction.mouseDown )
        Widget.bind( self.canvas, "<Button1-Motion>", self.interaction.mouseMotion )
        Widget.bind( self.canvas, "<Button1-ButtonRelease>", self.interaction.mouseUp )
        self.canvas.bind( "<Configure>", self.canvasResizeCallback )
        Widget.bind_all( self.b3, "<BackSpace>", self.deleteSelectedBox )
        
        # start the work now
        self.readData()        
        self.master.mainloop()



    def getScriptLocation( self ) :
        """
            This function determines the location of the pGQL script on the
            harddrive. This is necessary in order to load the icons correctly,
            when the current working directory of the shell where pGQL is called from
            is not equal to the location of pGQL on the harddrive or in other words
            pGQL is being started from a location outside of its own directory.
        """
        scriptLocation = os.path.abspath( __file__ )#the location of the file that contains this script 
        separator = os.sep
        directories = scriptLocation.split(separator)
        if len( directories ) > 1 :
            directories = directories[0:len( directories )-1 ] #exclude the script name itself
            if directories[ 0 ] == '' :
                scriptLocation = separator
            else :
                scriptLocation = ""
            for dir in directories :
                if dir != '' :
                    scriptLocation = scriptLocation +dir+separator
        if len( directories ) == 1 :
            scriptLocation = ""
        return scriptLocation



    def createGUI( self ) :
        """
            # The app's grid layout
            ####################################
            # <Logo>        # Toolbar     #
            # <Logo>        # QueryLabel  #
            # <Logo>        # <Canvas>    #
            # QueryControls # <Canvas>    #
            ####################################
            
            ####################################
            # Query Controls layout
            ####################################
            # rankSlider                       #
            # viewTextLabel                    #
            # viewLogRB     viewLinRB          #
            #                                  #
            # interpolationTextLabel           #
            # linearVis     bezierVis          #
            #                                  #
            # boxLevelTextLabel                #
            # front     back                   #
            ####################################
        """
        # start the grid layout
        self.grid( sticky='N'+'S'+'E'+'W' )
        self.master.columnconfigure( 0,weight=0 )
        self.master.columnconfigure( 1,weight=1 )
        self.master.rowconfigure( 0, weight=0 )
        self.master.rowconfigure( 1, weight=0 )
        self.master.rowconfigure( 2,weight=0 )   #the canvas' cell takes up any additional space from expansion
        self.master.rowconfigure( 3, weight=1 )
        self.winfo_toplevel().resizable( width=True, height=True )
                
        # add GQL logo
        self.gqlLogo = PhotoImage( file=self.scriptLocation+"icons/pGQL.gif" )
        self.logoHolder = Label( self.master, image=self.gqlLogo )
        self.logoHolder.grid( row=0, column=0, rowspan=3, sticky=NW )
                
        # start query controls layout
        self.queryControls = Frame( self.master, width=166, height=200 )
        self.queryControls.grid( row=3, column=0, sticky=NW )
        
        self.sliderVar=IntVar()
        self.rankSlider = Scale( self.queryControls, orient=HORIZONTAL, resolution=1, length=150, variable=self.sliderVar,\
                                 from_=0, to=0, label="Show top ranked", command=self.sliderCallback )
        self.rankSlider.grid( row=0, column=0, sticky=NW, pady=25 )
        self.rankSlider.set( 100 )
        
        self.visualizationOptionsTextLabel = Label( self.queryControls, text="Visualization options:" )
        self.visualizationOptionsTextLabel.grid(row=1,column=0,sticky=NW, padx=5)
        
        self.viewTextLabel = Label( self.queryControls, text="Scaling" )
        self.viewTextLabel.grid( row=2, column=0, sticky=NW, padx=15 )
        self.viewScaleVar = StringVar()
        self.viewScaleLinearRB = Radiobutton( self.queryControls, value="linear", variable=self.viewScaleVar,\
                                              text="linear",command=self.renderer.redisplay )
        self.viewScaleLinearRB.grid( row=3, column=0, sticky=NW, padx=15 )
        self.viewScaleMaxRB = Radiobutton( self.queryControls, value="max", variable=self.viewScaleVar, \
                                           text="maximize (affine)",command=self.renderer.redisplay )
        self.viewScaleMaxRB.grid( row=4, column=0, sticky=NW, padx=15 )
        self.viewScaleLinearRB.invoke()
        
        self.interpolationTextLabel = Label( self.queryControls, text="Interpolation" )
        self.interpolationTextLabel.grid( row=5, column=0, sticky=NW, padx=15 )
        self.interpolationVar = StringVar()
        self.linearVis = Radiobutton( self.queryControls, value="false", variable=self.interpolationVar,\
                                      text="linear",command=self.renderer.redisplay )
        self.linearVis.grid( row=6, column=0, sticky=NW, padx=15 )
        self.bezierVis = Radiobutton( self.queryControls, value="bezier", variable=self.interpolationVar,\
                                      text="bezier",command=self.renderer.redisplay )
        self.bezierVis.grid( row=7, column=0, sticky=NW, padx=15 )
        self.linearVis.invoke()

        self.queryModelTextLabel = Label( self.queryControls, text="Query Model" )
        self.queryModelTextLabel.grid( row=8, column=0, sticky=NW, padx=15 )
        self.queryModelVar = StringVar()
        self.deterministicModel = Radiobutton( self.queryControls, value="deterministic", variable=self.queryModelVar,\
                                      text="deterministic",command=self.queryModelCallback )
        self.deterministicModel.grid( row=9, column=0, sticky=NW, padx=15 )
        self.probabilisticModel = Radiobutton( self.queryControls, value="probabilistic", variable=self.queryModelVar,\
                                      text="probabilistic",command=self.queryModelCallback )
        self.probabilisticModel.grid( row=10, column=0, sticky=NW, padx=15 )
        self.probabilisticModel.invoke()

        
        #add query label
        self.queryLabel = Label( self.master, text="Query result from file <None>", bg="gray" )
        self.queryLabel.grid( row=1, column=1, sticky=NW )
        #add toolbar
        self.toolbar = Frame( self.master, width=200, height=20 )
        self.toolbar.grid( row=0, column=1, sticky=NW )
                
        self.openInactiveIcon = PhotoImage( file=self.scriptLocation+"icons/open.gif" )
        self.openIcon = PhotoImage( file=self.scriptLocation+"icons/open.gif" )
        self.b0 = Button( self.toolbar, image=self.openInactiveIcon, relief=RAISED, command=self.readData )
        self.b0.pack( side=LEFT )
        openButtonToolTip = ToolTip( self.b0 )
        openButtonToolTip.createToolTip( self.b0, "open" )
                
        self.saveInactiveIcon = PhotoImage( file=self.scriptLocation+"icons/saveInactive.gif" )
        self.saveIcon = PhotoImage( file=self.scriptLocation+"icons/save.gif" )
        self.b1 = Button( self.toolbar, image=self.saveInactiveIcon, relief=RAISED, command=self.saveData )
        self.b1.pack( side=LEFT )
        saveButtonToolTip = ToolTip( self.b1 )
        saveButtonToolTip.createToolTip( self.b1, "save" )

        self.saveCanvasInactiveIcon = PhotoImage( file=self.scriptLocation+"icons/saveCanvasInactive.gif" )
        self.saveCanvasIcon = PhotoImage( file=self.scriptLocation+"icons/saveCanvas.gif" )                
        self.b6 = Button( self.toolbar, image=self.saveCanvasInactiveIcon, relief=RAISED, command=self.saveCanvas )
        self.b6.pack( side=LEFT )
        saveCanvasButtonToolTip = ToolTip( self.b6 )
        saveCanvasButtonToolTip.createToolTip( self.b6, "save canvas as postscript" )
        
        self.gProfilerIcon = PhotoImage( file=self.scriptLocation+"icons/gProfiler.gif" )
        self.b10 = Button( self.toolbar, image=self.gProfilerIcon, relief=RAISED, command=self.submitTOgProfiler )
        self.b10.pack( side=LEFT )
        gProfilerButtonToolTip = ToolTip( self.b10 )
        gProfilerButtonToolTip.createToolTip( self.b10, "submit query result to gProfiler \nhttp://biit.cs.ut.ee/gprofiler/" )
                
        self.rectangleInactiveIcon = PhotoImage( file=self.scriptLocation+"icons/addBoxInactive.gif" )
        self.rectangleIcon = PhotoImage( file=self.scriptLocation+"icons/addBox.gif" )
        self.b2 = Button( self.toolbar, image=self.rectangleInactiveIcon, relief=RAISED, command=self.setDrawMode )
        self.b2.pack( side=LEFT )
        drawButtonToolTip = ToolTip( self.b2 )
        drawButtonToolTip.createToolTip( self.b2, "draw box" )
            
        self.moveInactiveIcon = PhotoImage( file=self.scriptLocation+"icons/moveBoxInactive.gif" )
        self.moveIcon = PhotoImage( file=self.scriptLocation+"icons/moveBox.gif" )
        self.b3 = Button( self.toolbar, image=self.moveInactiveIcon, relief=RAISED, command=self.setMoveMode )
        self.b3.pack( side=LEFT )
        moveButtonToolTip = ToolTip( self.b3 )
        moveButtonToolTip.createToolTip( self.b3, "select and move box" )
        
        self.resizeInactiveIcon = PhotoImage( file=self.scriptLocation+"icons/resizeBoxInactive.gif" )
        self.resizeIcon = PhotoImage( file=self.scriptLocation+"icons/resizeBox.gif" )
        self.b4 = Button( self.toolbar, image=self.resizeInactiveIcon, relief=RAISED, command=self.setResizeMode )
        self.b4.pack( side=LEFT )
        resizeButtonToolTip = ToolTip( self.b4 )
        resizeButtonToolTip.createToolTip( self.b4, "resize box" )
            
        self.deleteBoxIcon = PhotoImage( file=self.scriptLocation+"icons/deleteBox.gif" )    
        self.b9 = Button( self.toolbar, image = self.deleteBoxIcon, relief=RAISED, command = self.deleteSelectedBox )
        self.b9.pack( side=LEFT )
        deleteBoxButtonToolTip = ToolTip( self.b9 )
        deleteBoxButtonToolTip.createToolTip( self.b9, "delete selected box" )
        
        self.displayOrderInactiveIcon = PhotoImage( file=self.scriptLocation+"icons/displayOrderInactive.gif" )  
        self.displayOrderIcon = PhotoImage( file=self.scriptLocation+"icons/displayOrder.gif" )  
        self.b8 = Button( self.toolbar, image=self.displayOrderIcon, relief=RAISED, command=self.displayOrder )
        self.b8.pack( side=LEFT )
        displayOrderToolTip = ToolTip( self.b8 )
        displayOrderToolTip.createToolTip( self.b8, "change time box display level" )
        self.boxLevelVar = "front"
                
        self.quitInactiveIcon = PhotoImage( file=self.scriptLocation+"icons/quitInactive.gif" )
        self.quitIcon = PhotoImage( file=self.scriptLocation+"icons/quit.gif" )
        self.b5 = Button( self.toolbar, image=self.quitInactiveIcon, relief=RAISED, command=self.setQuitMode )
        self.b5.pack( side=LEFT )
        quitButtonToolTip = ToolTip( self.b5 )
        quitButtonToolTip.createToolTip( self.b5, "quit application" )
            
        self.helpInactiveIcon = PhotoImage( file=self.scriptLocation+"icons/helpInactive.gif" )
        self.helpIcon = PhotoImage( file=self.scriptLocation+"icons/help.gif" )
        self.b7 = Button( self.toolbar, image=self.helpInactiveIcon, relief=RAISED, command=self.help )
        self.b7.pack( side=RIGHT )
        quitButtonToolTip = ToolTip( self.b7 )
        quitButtonToolTip.createToolTip( self.b7, "help" )
    
    
    def displayOrder( self ) :
        if self.boxLevelVar == "front" :
            self.boxLevelVar = "back"
            self.b8.config( relief=SUNKEN, image=self.displayOrderInactiveIcon )
            self.renderer.adjustRenderingOrder()
        else :
            self.boxLevelVar = "front"
            self.b8.config( relief=RAISED, image=self.displayOrderIcon )
            self.renderer.adjustRenderingOrder()
    
    
    def canvasResizeCallback( self, event ) :
        if( self.width != event.width or self.height != event.height ) :
            self.width=event.width
            self.height=event.height
            self.renderer.redisplay()
    
    
    
    def deleteSelectedBox( self, event="" ) :
        if( self.interaction.tempBox != None and self.mode=="move" ) :
            for i in range( len( self.data.timeBoxes ) ) :
                if self.data.timeBoxes[ i ] == self.interaction.tempBox :
                    del self.data.timeBoxes[ i ]
                    self.renderer.deleteWidget( self.interaction.tempBox.getCanvasID() )
                    self.interaction.tempBox = None
                    self.data.recomputeHiddenBoxes()
                    self.interaction.queryChanged()
                    break            
            
            
            
    def setButtonAppearance( self ) :
        if self.mode == "no mode" :
            self.interaction.tempBox = None
            self.canvas.config( cursor="left_ptr" )
            self.setResizeHandles( False )
            for tb in self.data.timeBoxes :
                self.renderer.setOutline( tb.getCanvasID(), 1 )
            self.b0.config( relief=RAISED, image=self.openInactiveIcon )
            self.b1.config( relief=RAISED, image=self.saveInactiveIcon )
            self.b2.config( relief=RAISED, image=self.rectangleInactiveIcon )
            self.b3.config( relief=RAISED, image=self.moveInactiveIcon )
            self.b4.config( relief=RAISED, image=self.resizeInactiveIcon )
            self.b5.config( relief=RAISED, image=self.quitInactiveIcon )
            self.b6.config( relief=RAISED, image=self.saveCanvasInactiveIcon )
            self.b7.config( relief=RAISED, image=self.helpInactiveIcon )
        if self.mode == "open" :
            self.interaction.tempBox = None
            self.canvas.config( cursor="left_ptr" )
            self.setResizeHandles( False )
            for tb in self.data.timeBoxes :
                self.renderer.setOutline( tb.getCanvasID(), 1 )
            self.b0.config( relief=SUNKEN, image=self.openIcon )
            self.b1.config( relief=RAISED, image=self.saveInactiveIcon ) 
            self.b2.config( relief=RAISED, image=self.rectangleInactiveIcon )
            self.b3.config( relief=RAISED, image=self.moveInactiveIcon )
            self.b4.config( relief=RAISED, image=self.resizeInactiveIcon )
            self.b5.config( relief=RAISED, image=self.quitInactiveIcon )
            self.b6.config( relief=RAISED, image=self.saveCanvasInactiveIcon )
            self.b7.config( relief=RAISED, image=self.helpInactiveIcon )
        if self.mode == "save" :
            self.interaction.tempBox = None
            self.canvas.config( cursor="left_ptr" )
            self.setResizeHandles( False )
            self.b0.config( relief=RAISED, image=self.openInactiveIcon )
            self.b1.config( relief=SUNKEN, image=self.saveIcon )
            self.b2.config( relief=RAISED, image=self.rectangleInactiveIcon )
            self.b3.config( relief=RAISED, image=self.moveInactiveIcon )
            self.b4.config( relief=RAISED, image=self.resizeInactiveIcon )
            self.b5.config( relief=RAISED, image=self.quitInactiveIcon )
            self.b6.config( relief=RAISED, image=self.saveCanvasInactiveIcon )
            self.b7.config( relief=RAISED, image=self.helpInactiveIcon )
        if self.mode == "draw" :
            self.interaction.tempBox = None
            self.canvas.config( cursor="left_ptr" )
            self.setResizeHandles( False )
            for tb in self.data.timeBoxes :
                self.renderer.setOutline( tb.getCanvasID(), 1 )
            self.b0.config( relief=RAISED, image=self.openInactiveIcon )
            self.b1.config( relief=RAISED, image=self.saveInactiveIcon )
            self.b2.config( relief=SUNKEN, image=self.rectangleIcon )
            self.b3.config( relief=RAISED, image=self.moveInactiveIcon )
            self.b4.config( relief=RAISED, image=self.resizeInactiveIcon )
            self.b5.config( relief=RAISED, image=self.quitInactiveIcon )
            self.b6.config( relief=RAISED, image=self.saveCanvasInactiveIcon )
            self.b7.config( relief=RAISED, image=self.helpInactiveIcon )
        if self.mode == "move" :
            self.canvas.config( cursor = "hand1" )
            self.setResizeHandles( False )
            self.b0.config( relief=RAISED, image=self.openInactiveIcon )
            self.b1.config( relief=RAISED, image=self.saveInactiveIcon )
            self.b2.config( relief=RAISED, image=self.rectangleInactiveIcon )
            self.b3.config( relief=SUNKEN, image=self.moveIcon )
            self.b4.config( relief=RAISED, image=self.resizeInactiveIcon )
            self.b5.config( relief=RAISED, image=self.quitInactiveIcon )
            self.b6.config( relief=RAISED, image=self.saveCanvasInactiveIcon )
            self.b7.config( relief=RAISED, image=self.helpInactiveIcon )
        if self.mode == "resize" : 
            self.interaction.tempBox = None
            self.canvas.config( cursor="left_ptr" ) 
            for tb in self.data.timeBoxes :
                self.renderer.setOutline( tb.getCanvasID(), 1 )
            self.b0.config( relief=RAISED, image=self.openInactiveIcon )
            self.b1.config( relief=RAISED, image=self.saveInactiveIcon )
            self.b2.config( relief=RAISED, image=self.rectangleInactiveIcon )
            self.b3.config( relief=RAISED, image=self.moveInactiveIcon )
            self.b4.config( relief=SUNKEN, image=self.resizeIcon )
            self.b5.config( relief=RAISED, image=self.quitInactiveIcon )
            self.b6.config( relief=RAISED, image=self.saveCanvasInactiveIcon )
            self.b7.config( relief=RAISED, image=self.helpInactiveIcon )      
        if self.mode == "save canvas" :
            for tb in self.data.timeBoxes :
                self.renderer.setOutline( tb.getCanvasID(), 1 )
            self.b0.config( relief=RAISED, image=self.openInactiveIcon )
            self.b1.config( relief=RAISED, image=self.saveInactiveIcon )
            self.b2.config( relief=RAISED, image=self.rectangleInactiveIcon )
            self.b3.config( relief=RAISED, image=self.moveInactiveIcon )
            self.b4.config( relief=RAISED, image=self.resizeInactiveIcon )
            self.b5.config( relief=RAISED, image=self.quitInactiveIcon )
            self.b6.config( relief=SUNKEN, image=self.saveCanvasIcon )
            self.b7.config( relief=RAISED, image=self.helpInactiveIcon )
        if self.mode=="help" :
            for tb in self.data.timeBoxes :
                self.renderer.setOutline( tb.getCanvasID(), 1 )
            self.b0.config( relief=RAISED, image=self.openInactiveIcon )
            self.b1.config( relief=RAISED, image=self.saveInactiveIcon )
            self.b2.config( relief=RAISED, image=self.rectangleInactiveIcon )
            self.b3.config( relief=RAISED, image=self.moveInactiveIcon )
            self.b4.config( relief=RAISED, image=self.resizeInactiveIcon )
            self.b5.config( relief=RAISED, image=self.quitInactiveIcon )
            self.b6.config( relief=RAISED, image=self.saveCanvasInactiveIcon )
            self.b7.config( relief=SUNKEN, image=self.helpIcon )
        if self.mode=="quit" :
            for tb in self.data.timeBoxes :
                self.renderer.setOutline( tb.getCanvasID(), 1 )
            self.b0.config( relief=RAISED, image=self.openInactiveIcon )
            self.b1.config( relief=RAISED, image=self.saveInactiveIcon )
            self.b2.config( relief=RAISED, image=self.rectangleInactiveIcon )
            self.b3.config( relief=RAISED, image=self.moveInactiveIcon )
            self.b4.config( relief=RAISED, image=self.resizeInactiveIcon )
            self.b5.config( relief=SUNKEN, image=self.quitIcon )
            self.b6.config( relief=RAISED, image=self.saveCanvasInactiveIcon )
            self.b7.config( relief=RAISED, image=self.helpInactiveIcon )



    def setResizeHandles( self, show ) :
        if( show ) :     
            for box in self.data.timeBoxes :
                if not box.hidden :
                    coords = self.canvas.coords( box.getCanvasID() )
                    x0=self.renderer.inverseScaleFunction( coords[ 0 ], 0 )[ 0 ]
                    y0=self.renderer.inverseScaleFunction( 0, coords[ 1 ] )[ 1 ]
                    x1=self.renderer.inverseScaleFunction( coords[ 2 ], 0 )[ 0 ]
                    y1=self.renderer.inverseScaleFunction( 0, coords[ 3 ] )[ 1 ]
                
                    o1 = self.canvas.create_oval( coords[ 0 ]-3, coords[ 1 ]-3, coords[ 0 ]+3, coords[ 1 ]+3, fill="royalblue", tag="handle" )
                    h1=ResizeHandle( x0, y0, o1, box.getCanvasID(), "downLeft" )
                    box.handles.append( h1 )
                
                    o2 = self.canvas.create_oval( coords[ 0 ]-3, coords[ 3 ]-3, coords[ 0 ]+3, coords[ 3 ]+3, fill="royalblue", tag="handle" )
                    h2=ResizeHandle( x0, y1, o2, box.getCanvasID(), "upLeft" )
                    box.handles.append( h2 )
                 
                    o3 = self.canvas.create_oval( coords[ 2 ]-3, coords[ 3 ]-3, coords[ 2 ]+3, coords[ 3 ]+3, fill="royalblue", tag="handle" )
                    h3=ResizeHandle( x1, y1, o3, box.getCanvasID(), "upRight" )
                    box.handles.append( h3 )
                
                    o4 = self.canvas.create_oval( coords[ 2 ]-3, coords[ 1 ]-3, coords[ 2 ]+3, coords[ 1 ]+3, fill="royalblue", tag="handle" )
                    h4=ResizeHandle( x1, y0, o4, box.getCanvasID(), "downRight" )
                    box.handles.append( h4 )
                
        if( not show ) :
            for box in self.data.timeBoxes :
                if not box.hidden :
                    for handle in box.handles :
                        self.renderer.deleteWidget( handle.canvasID )
                        del handle
                box.handles = []
                
            
    
    def setQuitMode( self ) :
        """
            Quits the application
        """
        self.mode = "quit"
        self.setButtonAppearance()
        self.master.quit()



    def setDrawMode( self ) :
        """
            This (en-)disables the draw function and changes the look of its button.
        """
        if self.mode != "draw" :
            self.mode = "draw"
            self.setButtonAppearance()
        else :
            self.mode = "no mode"
            self.setButtonAppearance()


    
    def sliderCallback( self, scaleValue ) :
        # deterministic queries have no rank
        if self.deterministicQuery :
            return
        else :
            self.query.setRankCutoff( int(scaleValue) )
    
    
    def queryModelCallback( self ) :
        if self.queryModelVar.get() == "deterministic" :
            self.deterministicQuery = True
        else :
            self.deterministicQuery = False
        if self.dataFile != "" :    
            self.interaction.queryChanged()                    


    def setMoveMode( self ) :
        """
            This (en-)disables the move function and changes the look of the cursor over the canvas
        """
        if self.mode != "move" :
            self.mode = "move"
            self.setButtonAppearance()
        else :
            self.mode = "no mode"
            self.setButtonAppearance()
    
    

    def setResizeMode( self ) :
        """
            This (en-)disables the resize function and changes the look of the cursor over the canvas
        """
        if self.mode != "resize" :
            self.mode = "resize"
            self.setButtonAppearance()
            self.setResizeHandles( True )
        else :
            self.mode = "no mode"
            self.setButtonAppearance()
            self.setResizeHandles( False )
            

    
    def saveData( self ) :
        if( self.mode != "save" ) :
            self.mode = "save"
            self.setButtonAppearance()
        else :
            self.mode = "no mode"
            self.setButtonAppearance()

        fileName = tkFileDialog.asksaveasfilename( title="save query as", initialfile="queryResult.txt" )

        if( fileName ) :
            self.data.writeFile( fileName, self.interaction.gui.query )
            
        if( self.mode != "no mode" ) :
            self.mode = "no mode"
            self.setButtonAppearance()
        else :
            self.mode = "no mode"
            self.setButtonAppearance()
        return
    
    
    
    def saveCanvas( self ) :
        if( self.mode != "save canvas" ) :
            self.mode = "save canvas"
            self.setButtonAppearance()
        else :
            self.mode = "no mode"
            self.setButtonAppearance()
        filename = tkFileDialog.asksaveasfilename( title="save graphics as", initialfile="query.ps" )
        if( filename ) :
            self.canvas.postscript( file=filename, colormode="color" )
        if( self.mode != "no mode" ) :
            self.mode = "no mode"
            self.setButtonAppearance()
        else :
            self.mode = "no mode"
            self.setButtonAppearance()
        return
    
    
    
    def readData( self ) :
        """
            Reads the file from the users selection
        """
        if( self.mode != "open" ) :
            self.mode = "open"
            self.setButtonAppearance()
        else :
            self.mode = "no mode"
            self.setButtonAppearance ()
        self.dataFile = tkFileDialog.askopenfilename ( title="choose .txt data file", \
                    defaultextension=".txt", filetypes = [("Cage compatible", ".txt"), ("all",".*")] )
        if self.dataFile :
            if( extension( self.dataFile ) == "txt" ) :
                self.renderer.clearCanvas()
                self.data.timeBoxes=[]
                self.interaction.tempBox = None
                self.data.readFile( self.dataFile )
                self.renderer.display()
            else :
                print "sorry, we only read .txt files at the minute. Check our supported file formats at http://ghmm.org/gql"
        if( self.mode != "no mode" ) :
            self.mode = "no mode"
            self.setButtonAppearance()
        else :
            self.mode = "no mode"
            self.setButtonAppearance()
    
    
    
    def submitTOgProfiler( self ) :
        if self.b10.cget( 'relief' ) == 'RAISED' :
            self.b10.configure( relief=SUNKEN )
        if self.b10.cget( 'relief' ) == 'SUNKEN' :
            self.b10.configure( relief=RAISED )
        self.data.functionalAnalysisQuery(self.interaction.gui.query )
        return
    
    
    
    def help( self ) :
        if( self.mode != "help" ) :
            self.mode = "help"
            self.setButtonAppearance()
    
        tkMessageBox.showinfo(
                "Help",
                "\nThis is pGQL Version 0.1\n\nDocumentation is available here:\nhttp://ghmm.org/gql\n \
                \nAuthors:\nRuben B. Schilling, Ivan G. Costa, Alexander Schliep\n"
                )
        self.mode = "no mode"
        self.setButtonAppearance()
        return
#------------------------------------------------------------------------------------------------------
#******************************************************************************************************
#------------------------------------------------------------------------------------------------------



#------------------------------------------------------------------------------------------------------
#******************************************************************************************************
#------------------------------------------------------------------------------------------------------
#script execution begins here
if( __name__ == '__main__' ) :
    window = GUIMaster()
